require 'wechat_client/api'
require 'rest-client'
require 'monitor'

module WechatClient
  class WechatApiNotFoundError < NoMethodError
    require 'fuzzy_match'
    @@methods = WechatClient::Api.public_instance_methods(false)
    Finder = FuzzyMatch.new(@@methods)

    def initialize(name)
      fuzzy_method_name = Finder.find(name)
      error_msg = "undefined wechat api: #{name}"
      error_msg << ". Do you mean #{fuzzy_method_name}?" unless fuzzy_method_name.empty?
      super(error_msg)
    end
  end

  class ResponseCodeError < StandardError
    def initialize(error_json)
      super "#{error_json['errcode']}:#{error_json['errmsg']}"
    end
  end

  class Client
    extend MonitorMixin
    attr_reader :error, :url

    DEFAULT = {
      app_id:          nil,
      app_secret:      nil,
      serial_no:       nil,
      raise_flag:      false,  # 微信返回错误代码时raise
      verify_ssl:      true,   # RestClient::Request.execute 参数
      auto_cache:      true    # 是否使用自带的缓存方式
    }

    NEED_AUTO_CACHE = [:access_token, :jsapi_ticket]

    def initialize args
      @options = DEFAULT.merge(args.inject({}){|memo,(k,v)| memo[k.to_sym] = v; memo})
      @api = Api.new(options)
      @error = {}
    end

    GETTING_APIS = WechatClient::Apis::Getting.public_instance_methods(false)
    POSTING_APIS = WechatClient::Apis::Posting.public_instance_methods(false)

    GETTING_APIS.each do |api|
      define_method api do |params = {}, access_token = nil, &block|
        get(api, params, access_token, &block)
      end
    end

    POSTING_APIS.each do |api|
      define_method api do |post_data = {}, access_token = nil, &block|
        post(api, post_data, access_token, &block)
      end
    end

    def get(api_name, params = {}, at = nil, &block)
      Client.synchronize do
        @api.access_token = at
        response = request_handle(api_name, params) do |params|
          @url += params.to_query if params
          RestClient::Request.execute(method: :get, url: @url, verify_ssl: options.verify_ssl)
        end
        response_handle(api_name, response, options.raise_flag, &block)
      end
    end

    def post(api_name, post_data = {}, at = nil, &block)
      Client.synchronize do
        @api.access_token = at
        response = request_handle(api_name, post_data) do |post_data|
          RestClient::Request.execute(method: :post, url: @url, payload: post_data.to_json, verify_ssl: options.verify_ssl)
        end
        response_handle(api_name, response, options.raise_flag, &block)
      end
    end

    def options
      OpenStruct.new(@options)
    end

    private

    def cache_read api_name
      WechatClient.cache.exists("__wechat_client_#{api_name}__") ? \
        WechatClient.cache.read("__wechat_client_#{api_name}__") : yield(nil)
    end

    def cache_write api_name, json
      WechatClient.cache.write("__wechat_client_#{api_name}__", json)
    end

    def request_handle api_name, params, &block
      if @api.respond_to?(api_name)
        @url = @api.send(api_name)
      else
        raise WechatApiNotFoundError, api_name
      end

      return cache_read(api_name, &block) if options.auto_cache && NEED_AUTO_CACHE.include?(api_name.to_sym)
      yield params
    end

    def response_handle(api_name, response, raise_flag)
      # 重置 @error
      @error.clear
      content_type = response.respond_to?(:headers) ? response.headers[:content_type] : ""

      # 判断返回类型, 已知 json 和 image 两种
      data = if content_type.include?("json") || content_type.include?("text")
        json = JSON.parse(response.body.gsub(/[\u0000-\u001f]+/, ''))
        if json["errcode"] && json["errcode"].to_i != 0
          WechatClient.logger.error(json)
          raise(ResponseCodeError, json) if raise_flag
          @error = json
        else
          WechatClient.logger.info("WechatClientLog: api-#{api_name} return #{json}")
          cache_write(api_name, json)if options.auto_cache && NEED_AUTO_CACHE.include?(api_name.to_sym)
        end
        json
      elsif content_type.include? "image"
        tmp = Tempfile.new('tmp')
        tmp.binmode
        tmp.write(response)
        tmp.close
        tmp
      else
        response
      end

      if block_given? && @error.empty?
        yield(data)
      else
        data
      end
    end
  end
end
